###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from ..database.objdb import ObjDbBase, ObjDbClient
from ..database.objdb_api import UseDatabase
from ..database.tsdb import TsDbClient
from ..database.tsdb_api import TsDbUseDatabase
from ..depgraph.graph_api import UseGraph

import contextlib
import configparser
import getpass
import logging
import os

__all__ = ["load_system_configuration", "OnyxInit", "OnyxStartup"]

# --- this is the global instance of ConfigParser that is used to store
#     configuration parameters loaded from the onyx_config.ini file
__config = None

logger = logging.getLogger(__name__)


###############################################################################
class ConfigParserWithLists(configparser.ConfigParser):
    """
    A ConfigParser subclass extended to support list items represented as
    follows:

    [section]
    list = value1, value2, ....
    """
    # -------------------------------------------------------------------------
    def getlist(self, section, option):
        items = self.get(section, option).splitlines()
        return [item.strip() for item in items if item != ""]


# -----------------------------------------------------------------------------
def load_system_configuration(filenames=None):
    """
    Description:
        Load configuration options from one or more INI files. Return the
        cached value if called more than once (in that case, filenames will be
        ignored).
    Inputs:
        filenames - a list of possible config file paths. By default, looks for
                    a onyx_config.ini in the following folders (stops at the
                    first match):
                        $ONYXPATH
                        $HOME
                        $USERPROFILE
                        .
    Returns:
        An instance of configparser.ConfigParser
    """
    global __config

    if __config is None:
        __config = ConfigParserWithLists()

        if filenames is None:
            filepath = os.getenv("ONYXPATH",
                                 os.getenv("HOME",
                                           os.getenv("USERPROFILE", "./")))
            filenames = [os.path.join(filepath, "onyx_config.ini")]

        files = __config.read(filenames)

        if not len(files):
            logger.warning("couldn't find any of the specified "
                           "config files: {0!s}".format(filenames))

    return __config


# -----------------------------------------------------------------------------
def OnyxInit(objdb=None, tsdb=None, user=None, host=None, configfile=None):
    """
    Description:
        Activate onyx databases and graph using the respective context
        managers.
    Inputs:
        objdb - an instance of an ObjDb client or the name of a valid ObjDb
        tsdb  - an instance of an TsDb client or the name of a valid TsDb
        user  - the database user
        host  - the database host
        configfile - the full path of the config.ini file
    Returns:
        A stack of context managers.
    """
    config = load_system_configuration(configfile)

    objdb = objdb or config.get("database", "objdb", fallback="ProdDb")
    tsdb = tsdb or config.get("database", "tsdb", fallback="TsDb")
    user = user or config.get("database", "user", fallback=getpass.getuser())
    host = host or config.get("database", "host", fallback=None)

    if isinstance(objdb, ObjDbBase):
        objdb_clt = objdb
    else:
        objdb_clt = ObjDbClient(objdb, user, host)

    if isinstance(tsdb, TsDbClient):
        tsdb_clt = tsdb
    else:
        tsdb_clt = TsDbClient(tsdb, user, host)

    stack = contextlib.ExitStack()
    stack.enter_context(UseDatabase(objdb_clt))
    stack.enter_context(TsDbUseDatabase(tsdb_clt))
    stack.enter_context(UseGraph())

    return stack


# -----------------------------------------------------------------------------
def OnyxStartup(objdb=None, tsdb=None, user=None, host=None, configfile=None):
    """
    Description:
        Load onyx environment and open connections to backends.
    Typical Usage:
        In an interactive shell:
            globals().update(OnyxStartup())
    """
    import onyx.core

    stack = OnyxInit(objdb, tsdb, user, host, configfile)
    stack.__enter__()
    logger.info("Onyx has been fired up... Good luck!!!")

    return {key: value
            for key, value in onyx.core.__dict__.items()
            if not key.startswith("__")}



###############################################################################
#
#   Copyright: (c) 2017 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

import logging
import logging.config
import json

__all__ = ["setup_logging_from_json_config"]

DEFAULT_LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "simple": {
            "format": "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        }
    },

    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "level": "DEBUG",
            "formatter": "simple",
            "stream": "ext://sys.stdout"
        },
    },

    "root": {
        "level": "INFO",
        "handlers": ["console"]
    },
}


# -----------------------------------------------------------------------------
def setup_logging_from_json_config(config_file=None):
    """
    Description:
        Load configuration options for logging from a json file.
    Inputs:
        config_file - a logging configuration file in json format.
    Returns:
        None
    """
    if config_file is None:
        config = DEFAULT_LOGGING_CONFIG
    else:
        try:
            with open(config_file, "rt") as fin:
                config = json.load(fin)
        except FileNotFoundError:
            raise FileNotFoundError("logging configuration "
                "file {0:s} not found or readable".format(config_file))

    logging.config.dictConfig(config)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

import numpy as np
import collections
import bisect

__all__ = ["GCurve", "Knot", "CurveError"]


###############################################################################
class CurveError(Exception):
    """
    Base class for all Curve exceptions.
    """
    pass


###############################################################################
class Knot(object):
    """
    Class representing a generic curve knot as a pair (date, value).
    """
    __slots__ = ("date", "value")

    # -------------------------------------------------------------------------
    def __init__(self, d, v):
        self.date = d
        self.value = v

    # -------------------------------------------------------------------------
    def __repr__(self):
        cls = self.__class__
        return "{.__name__:s}({.date!r},{.value!r})".format(cls, self, self)

    # -------------------------------------------------------------------------
    def __str__(self):
        if hasattr(self.value, "__str__"):
            return "{0!s}, {1!s}".format(self.date, self.value)
        else:
            return "{0!s}, unprintable value...".format(self.date)

    # -------------------------------------------------------------------------
    #  methods for knot comparison (date based)
    def __lt__(self, other):
        return self.date < other.date

    def __le__(self, other):
        return self.date <= other.date

    def __gt__(self, other):
        return self.date > other.date

    def __ge__(self, other):
        return self.date >= other.date

    # -------------------------------------------------------------------------
    #  quality/inquality are based on both date and value
    def __eq__(self, other):
        return self.date == other.date and self.value == other.value

    def __ne__(self, other):
        return self.date != other.date or self.value != other.value


###############################################################################
class GCurve(object):
    """
    Base curve class implementing the core functionality. The internal
    representation is optimized for fast lookup (ln(N)) and quick access to
    dates and values. Inserts and deletes are slow.
    """
    # --- this is the default datatype for a GCurve
    dtype = np.object

    # -------------------------------------------------------------------------
    def __init__(self, dates=None, values=None, dtype=None):
        """
        Description:
            Create a curve from a list of dates and a list of generic values.
            Raise a CurveError exception if dates and values don't have same
            length or if dates contains any duplicates.
            Curve knots (date, value pairs) are sorted automatically.
        Inputs:
            dates  - list/array of Dates
            values - list/array of values
            dtype  - optional, data type for the ndarray of values
        """
        if dates is None: dates = []    # analysis:ignore
        if values is None: values = []  # analysis:ignore

        # --- number of items in the curve
        nknots = len(dates)

        if nknots != len(values):
            raise CurveError("Input dates and values are of different length")

        # --- check whether there are any duplicates in dates
        if nknots != len(set(dates)):
            duplicates = {x: y for x, y
                          in collections.Counter(dates).items() if y > 1}
            raise CurveError("Creating curve with duplicate "
                             "dates: {0:s}".format(duplicates.__repr__()))

        # --- sort list of knots by date
        data = sorted(zip(dates, values), key=lambda knot: knot[0])

        # --- a curve is stored as np.array for dates, one for values, and the
        #     data type used for values
        self.dtype = dtype or self.dtype
        self.dates = np.array([d for (d, v) in data], np.object)
        self.values = np.array([v for (d, v) in data], dtype)

    # -------------------------------------------------------------------------
    @classmethod
    def create_raw(cls, dates, values, dtype=None):
        """
        Description:
            Class method to create a curve bypassing all the checks performed
            during standard initialization.
        Inputs:
            dates  - list/array of Dates
            values - list/array of values
            dtype  - data type for the values (default is np.object for
                     a GCurve)
        """
        crv = cls.__new__(cls)
        crv.dtype = dtype or cls.dtype
        crv.dates = np.array(dates, dtype=np.object)
        crv.values = np.array(values, dtype=dtype)
        return crv

    # -------------------------------------------------------------------------
    def __deepcopy__(self, memo):
        clone = self.__class__.__new__(self.__class__)
        clone.dates = self.dates.copy()
        clone.values = self.values.copy()
        clone.dtype = self.dtype
        memo[id(self)] = clone
        return clone

    # -------------------------------------------------------------------------
    #  methods for curve comparison
    def __eq__(self, other):
        match = np.all(self.dates == other.dates)
        match = match and np.all(self.values == other.values)
        return match

    def __ne__(self, other):
        differ = np.any(self.dates != other.dates)
        differ = differ or np.any(self.values != other.values)
        return differ

    # -------------------------------------------------------------------------
    #  support serialization
    def __getstate__(self):
        return self.dates, self.values, self.dtype

    def __setstate__(self, state):
        self.dates, self.values, self.dtype = state

    # -------------------------------------------------------------------------
    def __repr__(self):
        cls = self.__class__.__name__
        dts = self.dates.tolist()
        vls = self.values.tolist()
        return "{0:s}(dates={1!r}, values={2!r})".format(cls, dts, vls)

    # -------------------------------------------------------------------------
    def __str__(self):
        return "\n".join([Knot(d, v).__str__()
                          for d, v in zip(self.dates, self.values)])

    # -------------------------------------------------------------------------
    #  GCurve is an iterable container type: implement some relevant methods
    def __len__(self):
        return len(self.dates)

    def __iter__(self):
        return zip(self.dates, self.values)

    def __getitem__(self, date):
        idx = self.dates.searchsorted(date)
        if self.dates[idx] == date:
            return self.values[idx]
        else:
            raise IndexError

    def __setitem__(self, date, val):
        idx = self.dates.searchsorted(date)
        if idx == len(self.dates):
            self.dates = np.append(self.dates, date)
            self.values = np.append(self.values, val)
        elif self.dates[idx] == date:
            self.values[idx] = val
        else:
            self.dates = np.insert(self.dates, idx, date)
            self.values = np.insert(self.values, idx, val)

    def __delitem__(self, date):
        idx = self.dates.searchsorted(date)
        if self.dates[idx] == date:
            self.dates = np.delete(self.dates, idx)
            self.values = np.delete(self.values, idx)
        else:
            raise IndexError

    def __contains__(self, date):
        dts = self.dates
        if not len(dts):
            return False
        if date < dts[0] or date > dts[-1]:
            return False
        if dts[dts.searchsorted(date)] == date:
            return True
        else:
            return False

    # -------------------------------------------------------------------------
    @property
    def ordinals(self):
        """
        Return an array of dates in numeric format.
        """
        return np.array([d.ordinal for d in self.dates], np.float64)

    # -------------------------------------------------------------------------
    @property
    def front(self):
        """
        Description:
            Return the first knot of the curve.
        Returns:
            A knot.
        """
        return Knot(self.dates[0], self.values[0])

    @property
    def back(self):
        """
        Description:
            Return the last knot of the curve.
        Returns:
            A knot.
        """
        return Knot(self.dates[-1], self.values[-1])

    # -------------------------------------------------------------------------
    def crop(self, start=None, end=None):
        """
        Description:
            Return a sub-curve with knots that fall within the specified range.
            If start date or end date are not set, use the current extremes of
            the curve.
        Inputs:
            start - the start of the range (included)
            end   - the   end of the range (included)
        Return:
            A sub-curve of the original curve.
        """
        if not len(self.dates):
            return self

        # --- store dates in local variable for faster access
        dts = self.dates

        if start is None:
            i = 0
        else:
            i = bisect.bisect_left(dts, start)

        if end is None:
            j = -1
        else:
            j = bisect.bisect_right(dts, end, i)

        return self.create_raw(dts[i:j], self.values[i:j])



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

import collections

__all__ = ["Structure", "StructureError"]


###############################################################################
class StructureError(Exception):
    """
    Base class for all Structure exceptions.
    """
    pass


###############################################################################
class Structure(collections.UserDict):
    """
    A Structure is a dictionary that only accepts numerical values and
    implements algebraic manipulations such as:
        - addition and subtraction of structures item by item
        - additiona and multiplication by a scalar
    """
    # -------------------------------------------------------------------------
    def __setitem__(self, key, value):
        try:
            value + 0.0
        except TypeError:
            raise StructureError("Incorrect Structure value for "
                                 "pair {0!s}: {1!s}".format(key, value))
        super().__setitem__(key, value)

    # -------------------------------------------------------------------------
    #  addition methods

    def __iadd__(self, other):
        if isinstance(other, Structure):
            for key, value in other.items():
                self[key] = self.get(key, 0.0) + value
        else:
            for key in self:
                self[key] += other
        return self

    def __add__(self, other):
        struct = self.copy()
        if isinstance(other, Structure):
            for key, value in other.items():
                struct[key] = struct.get(key, 0.0) + value
        else:
            for key in struct:
                struct[key] += other
        return struct

    def __radd__(self, scalar):
        struct = self.copy()
        for key in struct:
            struct[key] += scalar
        return struct

    # -------------------------------------------------------------------------
    #  subtraction methods

    def __isub__(self, other):
        if isinstance(other, Structure):
            for key, value in other.items():
                self[key] = self.get(key, 0.0) - value
        else:
            for key in self:
                self[key] -= other
        return self

    def __sub__(self, other):
        struct = self.copy()
        if isinstance(other, Structure):
            for key, value in other.items():
                struct[key] = struct.get(key, 0.0) - value
        else:
            for key in struct:
                struct[key] -= other
        return struct

    def __rsub__(self, scalar):
        struct = self.copy()
        for key in struct:
            struct[key] = scalar - struct[key]
        return struct

    # -------------------------------------------------------------------------
    #  multiplication methods

    def __imul__(self, scalar):
        if isinstance(scalar, Structure):
            raise StructureError("Multiplication of "
                                 "two Structures is not supported")
        try:
            scalar + 0.0
        except TypeError:
            raise StructureError("Multiplication by "
                                 "non-numeric scalar is not supported")
        for key in self:
            self[key] *= scalar
        return self

    def __mul__(self, scalar):
        if isinstance(scalar, Structure):
            raise StructureError("Multiplication of "
                                 "two Structures is not supported")
        try:
            scalar + 0.0
        except TypeError:
            raise StructureError("Multiplication by "
                                 "non-numeric scalar is not supported")
        struct = self.copy()
        for key in struct:
            struct[key] *= scalar
        return struct

    def __rmul__(self, scalar):
        try:
            scalar + 0.0
        except TypeError:
            raise StructureError("Multiplication by "
                                 "non-numeric scalar is not supported")
        struct = self.copy()
        for key in struct:
            struct[key] *= scalar
        return struct

    # -------------------------------------------------------------------------
    def __deepcopy__(self, memo):
        clone = self.copy()
        memo[id(self)] = clone
        return clone

    # -------------------------------------------------------------------------
    def drop_zeros(self, ndigits=None):
        """
        Description:
            Drop items with zero value (up to a given number of digits) in
            place.
        Inputs:
            ndigits - if set, values are rounded to the required number of
                      digits before testing for equality to zero.
        Returns:
            The clensed Structure.
        """
        if ndigits is None:
            self.data = {k: v for k, v in self.items() if v != 0.0}
        else:
            self.data = {k: v for k, v in
                         self.items() if round(v, ndigits) != 0.0}

        return self



###############################################################################
#
#   Copyright: (c) 2017 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

import subprocess
import sys
import logging
import argh


# -----------------------------------------------------------------------------
def run_forever(script=None, executable=None, other=None):
    """
    Description:
        Run a executable or python script and automatically restart it on
        failure.
    Inputs:
        script     - if specified, this is the python script that will be
                     executed in a subprocess
        executable - if script is specified, this is the python executable used
                     to run the script (default is sys.executable), otherwise
                     this is the program that will be executed in a subprocess
        other      - a string with arguments to be passed to the subprocess
    Returns:
        None
    """
    if executable is None and script is None:
        raise RuntimeError("Both executable and script are unspecified")

    executable = executable or sys.executable

    if script is None:
        args = [executable]
    else:
        args = [executable, script]

    if other is not None:
        args.append(other)

    # --- configure logger for the run_forever script
    config = {
        "level": logging.INFO,
        "format": "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    }
    logging.basicConfig(**config)
    logger = logging.getLogger(__name__)

    if sys.platform == "win32":
        # Don't display the Windows GPF dialog if the invoked program dies.
        # See comp.os.ms-windows.programmer.win32
        #  How to suppress crash notification dialog?, Jan 14,2004 -
	 #     Raymond Chen's response [1]   
        import ctypes
        SEM_NOGPFAULTERRORBOX = 0x0002 # From MSDN
        ctypes.windll.kernel32.SetErrorMode(SEM_NOGPFAULTERRORBOX);
        subprocess_flags = 0x8000000
    else:
        subprocess_flags = 0

    while True:
        try:
            subprocess.check_call(
                args, shell=False, creationflags=subprocess_flags)
            logger.info("subprocess completed without errors, restarting...")
        except subprocess.CalledProcessError as err:
            logger.error("subprocess died with error: {0!s}".format(err))


# -----------------------------------------------------------------------------
def main():
    argh.dispatch_command(run_forever)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from .datatypes.date import *  # analysis:ignore
from .datatypes.rdate import *  # analysis:ignore
from .datatypes.holiday_cal import *  # analysis:ignore
from .datatypes.gcurve import *  # analysis:ignore
from .datatypes.curve import *  # analysis:ignore
from .datatypes.hlocv import *  # analysis:ignore
from .datatypes.table import *  # analysis:ignore
from .datatypes.structure import *  # analysis:ignore

from .database.objdb import *  # analysis:ignore
from .database.objdb_api import *  # analysis:ignore
from .database.ufo_base import *  # analysis:ignore
from .database.ufo_fields import *  # analysis:ignore
from .database.tsdb import *  # analysis:ignore
from .database.tsdb_api import *  # analysis:ignore

from .depgraph.graph import *  # analysis:ignore
from .depgraph.graph_api import *  # analysis:ignore
from .depgraph.graph_scopes import *  # analysis:ignore
from .depgraph.ufo_archivable import *  # analysis:ignore
from .depgraph.ufo_functions import *  # analysis:ignore

from .libs.date_fns import *  # analysis:ignore
from .libs.curve_fns import *  # analysis:ignore
from .libs.table_fns import *  # analysis:ignore

from .utils.base36 import *  # analysis:ignore
from .utils.startup import *  # analysis:ignore
from .utils.unittest import *  # analysis:ignore
from .utils.logging import *  # analysis:ignore



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from ..datatypes.date import Date, DateError, L2M
from ..datatypes.rdate import RDate
from ..datatypes.holiday_cal import HolidayCalendar
from ..database.objdb_api import GetObj

__all__ = [
    "LYY2Date",
    "Date2LYY",
    "Date2QQYY",
    "CalcTerm",
    "DateOffset",
    "DateRange",
    "CountBizDays",
]

# --- auxiliary local variables (not exported)
M2L = dict(zip(L2M.values(), L2M.keys()))
QUARTER_BY_MTH = [1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4]


# -----------------------------------------------------------------------------
def LYY2Date(lyy):
    """
    Description:
        Convert a Nymex LYY string (used for futures contracts as in F09) to
        a Date. The LYY string is assumed to correspond to the first day in the
        corresponding month.
    Inputs:
        lyy - a string defining a valid Nymex future contract.
    Returns:
        A Date.
    """
    try:
        m = L2M[lyy[0]]
        y = int(lyy[1:])
        y = 2000 + y if y < 70 else 1900 + y
        return Date(y, m, 1)
    except KeyError:
        raise DateError("Unrecognized LYY string: {0:s}".format(lyy))


# -----------------------------------------------------------------------------
def Date2LYY(date):
    """
    Description:
        Convert a Date to a Nymex LYY string (used for futures contracts
        as in F09). Conversion is done ignoring the actual day in the month.
    Inputs:
        date - a Date for which the LYY Nymex future contract is returned.
    Returns:
        A string with the LYY code.
    """
    if date < Date.low_date() or date > Date.high_date():
        raise DateError("Date outside valid range: {0:s}".format(date))

    m = date.month
    y = date.year
    yy = y - 2000 if y > 1999 else y - 1900

    return "{0:1s}{1:02d}".format(M2L[m], yy)


# -----------------------------------------------------------------------------
def Date2QQYY(date):
    """
    Description:
        Convert a Date to a string for the corresponding quarter (as in Q311).
        Conversion is done ignoring the actual day in the quarter.
    Inputs:
        date - a Date for which the corresponding quarter is returned.
    Returns:
        A string with the QQYY code.
    """
    if date < Date.low_date() or date > Date.high_date():
        raise DateError("Date outside valid range: {0:s}".format(date))

    m = date.month - 1
    y = date.year
    yy = y - 2000 if y > 1999 else y - 1900

    return "Q{0:1d}{1:02d}".format(QUARTER_BY_MTH[m], yy)


# -----------------------------------------------------------------------------
def CalcTerm(start, end, base=365):
    """
    Description:
        Return the number of years between start and end date assuming that
        each year has a number of days as specified by base. Term is negative
        if start > end.
    Inputs:
        start - the initial date of the term
        end   - the final   date of the term
        base  - number of days in a calendar year
    Returns:
        A float.
    """
    return (end - start).days / float(base)


# -----------------------------------------------------------------------------
def DateOffset(date, rule, calendar=None):
    """
    Description:
        Applies the specified rule (a string, as defined below) to the
        specified date to obtain a new shifted date.
    Inputs:
        date     - date on which the date-rule is to be applied
        rule     - a tag specifying shift relative to input date
        calendar - a holiday calendar name used to identify business days
    Rule definitions:
        d = add calendar day
        b = add business day
        w = add calendar week
        m = add calendar month
        y = add calendar year
        c = go to the required day in the month
        e = go to end of month (ignores num)
        J = go to first calendar day of month (ignores num)
        M = go to closest Monday as specified by num
        W = go to closest Wednesday as specified by num
        F = go to closest Friday as specified by num
        q = go to beginning of the quarter (ignores num)
        Q = go to end of the quarter (ignores num)
        A = go to beginning of the year (ignores num)
        E = go to end of the year (ignores num)
    Returns:
        A Date.
    Examples:
        EndOfMonth = DateOffset(Date.today(), "+e")
        FirstBizDayOf2000 = DateOffset(Date(2000, 1, 1), "+b")
    """
    if calendar is None or isinstance(calendar, HolidayCalendar):
        pass
    elif isinstance(calendar, str):
        calendar = GetObj(calendar)
    else:
        raise ValueError("calendar must be either a string or "
                         "an instance of HolidayCalendar (a {0:s} "
                         "was passed instead).".format(type(calendar)))

    return date + RDate(rule, calendar)


# -----------------------------------------------------------------------------
def DateRange(start, end, rule, calendar=None):
    """
    Description:
        Iterator that yields a series of dates between start date and
        end date according to an increment rule (start date is included).
    Inputs:
        start    - the first date in the range, included
        end      - the last  date in the range, included or not, depending on
                   the rule
        rule     - a date rule (see RDate)
        calendar - a holiday calendar used to identify business days
    Yields:
        A range of Dates.
    """
    if calendar is None or isinstance(calendar, HolidayCalendar):
        pass
    elif isinstance(calendar, str):
        calendar = GetObj(calendar)
    else:
        raise ValueError("calendar must be either a string or "
                         "an instance of HolidayCalendar (a {0:s} "
                         "was passed instead).".format(type(calendar)))

    d = start
    rd = RDate(rule, calendar)
    while d <= end:
        yield d
        d = d + rd


# -----------------------------------------------------------------------------
def CountBizDays(start, end, calendar=None):
    """
    Description:
        Return the number of business days between start and end date according
        to the holiday calendar.
        NB: if start date == end date this function will return 1.
    Inputs:
        start    - the initial date of the range
        end      - the last  date in the range
        calendar - a holiday calendar used to identify business days
    Returns:
        An integer.
    """
    if calendar is None or isinstance(calendar, HolidayCalendar):
        pass
    elif isinstance(calendar, str):
        calendar = GetObj(calendar)
    else:
        raise ValueError("calendar must be either a string or "
                         "an instance of HolidayCalendar (a {0:s} "
                         "was passed instead).".format(type(calendar)))

    # --- first push start date forward to the nearest business day
    start = start + RDate("+0b", calendar)

    # --- then sum all business days in the range
    return sum((1 for d in DateRange(start, end, "+1b", calendar)))



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

# --- global instances of the active database clients (used by the APIs)
obj_clt = None
ts_clt = None

# --- For each object retrieved from database we want to have one and only one
#     instance per session, shared by each client and by the dependency graph.
#     This, not only improves performance (at the cost of heavy memory usage),
#     but, more importantly, ensures the proper functioning of the dependency
#     graph.
obj_instances = dict()



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from ..datatypes.date import Date
from .ufo_base import SPECIAL, custom_encoder, custom_decoder

from .. import database as onyx_db

import psycopg2 as psql
import psycopg2.extras as psql_extras
import psycopg2.extensions as psql_ext
import psycopg2.errorcodes as psql_err
import pickle
import json
import abc

__all__ = ["ObjNotFound", "ObjDbError", "ObjExists", "ObjDbClient"]

# --- here we define standard query strings and error messages

QUERY_CLS_INSERT = "INSERT INTO ClassDefinitions VALUES (%s,%s);"
QUERY_CLS_GET = """
SELECT ClassDef FROM ClassDefinitions WHERE ObjType=%s;"""
QUERY_CLS_DELETE = """
DELETE FROM ClassDefinitions row WHERE row.ObjType=%s AND NOT
EXISTS (SELECT 1 FROM Objects WHERE ObjType=row.ObjType);"""

QUERY_OBJ_INSERT = "INSERT INTO Objects VALUES (%s,%s,%s,%s,%s,%s,%s);"
QUERY_OBJ_GET = "SELECT * FROM Objects WHERE Name=%s;"
QUERY_OBJ_UPDATE = """
UPDATE Objects
SET Version=%s, ChangedBy=%s, LastUpdated=%s, Data=%s
WHERE Name=%s AND Version=%s;"""
QUERY_OBJ_DELETE = "SELECT delete_obj(%s, %s) AS deleted;"
QUERY_OBJ_VERSION = "SELECT Version FROM Objects WHERE Name=%s LIMIT 1;"
QUERY_OBJ_EXISTS = """
SELECT EXISTS (SELECT 1 FROM Objects WHERE Name=%s) AS "exists";"""

QUERY_ARC_INSERT = "INSERT INTO Archive VALUES (%s,%s,%s,%s,%s,%s);"
QUERY_ARC_UPDATE = """
UPDATE Archive
SET ChangedBy=%s, TimeStamp=%s, Value=%s
WHERE Name=%s AND Attribute=%s AND Date=%s;"""
QUERY_ARC_GET = """
SELECT Value, Date FROM Archive
WHERE Name=%s AND Attribute=%s AND Date IN
    (SELECT MAX(Date) FROM Archive
     WHERE Name=%s AND Attribute=%s AND Date<=%s);"""
QUERY_ARC_GET_STRICT = """
SELECT Value, Date FROM Archive
WHERE Name=%s AND Attribute=%s AND Date=%s;"""

MSG_RELOAD = """
Instance of {0:s} is older (v.{1:d}) than the most recent version in database
(v.{2:d}): reload object from database before updating"""


###############################################################################
class ObjNotFound(Exception):
    pass


###############################################################################
class ObjDbError(Exception):
    pass


###############################################################################
class ObjExists(ObjDbError):
    pass


###############################################################################
class dummy_conn(object):
    """
    This is a dummy connection class that implements placeholders methods
    used to support transactions. It is used by the base and dummy clients
    and should not be used by production implementations.
    """
    def set_isolation_level(self, *args):
        pass

    def tpc_begin(self, *args):
        pass

    def tpc_commit(self, *args):
        pass

    def tpc_rollback(self, *args):
        pass


###############################################################################
class ObjDbBase(object):
    """
    Base class for ObjDb clients.
    """
    # -------------------------------------------------------------------------
    def __init__(self, database, user, host=None):
        self.dbname = database
        self.user = user
        self.host = host

        # --- lookup dictionary for class definitions
        self.class_defs = {}

        # --- the connection to the db-backend
        self.conn = dummy_conn()

        # --- store current transaction name
        self.transaction = None

    # -------------------------------------------------------------------------
    def class_lookup(self, cls_name):
        """
        Lookup class by class name.
        """
        return self.class_defs[cls_name]

    # -------------------------------------------------------------------------
    #  implements methods to retrieve, add and remove items from db-cache
    def __getitem__(self, name):
        return onyx_db.obj_instances[name]

    def __setitem__(self, name, instance):
        onyx_db.obj_instances[name] = instance

    def __delitem__(self, name):
        try:
            del onyx_db.obj_instances[name]
        except KeyError:
            pass

    # -------------------------------------------------------------------------
    #  method to test if an object exists in database. the base class
    #  implementation only checks if a reference exsists without testing the
    #  backend.
    def __contains__(self, name):
        return name in onyx_db.obj_instances

    # -------------------------------------------------------------------------
    @abc.abstractmethod
    def initialize(self):
        raise NotImplementedError()

    # -------------------------------------------------------------------------
    @abc.abstractmethod
    def restart(self):
        raise NotImplementedError()

    # -------------------------------------------------------------------------
    @abc.abstractmethod
    def close(self):
        raise NotImplementedError()

    # -------------------------------------------------------------------------
    @abc.abstractmethod
    def add(self, obj):
        raise NotImplementedError()

    # -------------------------------------------------------------------------
    @abc.abstractmethod
    def get(self, name, refresh=False):
        raise NotImplementedError()

    # -------------------------------------------------------------------------
    @abc.abstractmethod
    def update(self, obj):
        raise NotImplementedError()

    # -------------------------------------------------------------------------
    @abc.abstractmethod
    def delete(self, obj):
        raise NotImplementedError()

    # -------------------------------------------------------------------------
    @abc.abstractmethod
    def set_dated(self, name, attr, date, value, overwrite=False):
        raise NotImplementedError()

    # -------------------------------------------------------------------------
    @abc.abstractmethod
    def get_dated(self, name, attr, date, strict=False):
        raise NotImplementedError()


# -----------------------------------------------------------------------------
def typecast_to_Date(value, curs):
    if value is None:
        return None
    else:
        # --- maybe we should use a regexp like r"^.*'(.*)'.*$" ???
        return Date.parse(value)


###############################################################################
class ObjDbClient(ObjDbBase):
    core_tables = ("ClassDefinitions", "Objects", "Archive")

    # -------------------------------------------------------------------------
    def __init__(self, database, user, host=None, check=True):
        """
        Description:
            ObjDb database, client class. It exposes a minimal api.
        """
        super().__init__(database, user, host)

        # --- open database connection
        self.conn = psql.connect(host=host, database=database, user=user,
                                 cursor_factory=psql_extras.NamedTupleCursor)
        self.conn.set_isolation_level(psql_ext.ISOLATION_LEVEL_AUTOCOMMIT)

        # --- register typecast to Date for date, timpestamp, timestamptz
        curs = self.conn.cursor()
        curs.execute("SELECT NULL::date, NULL::timestamp, NULL::timestamptz;")
        oids = tuple([col.type_code for col in curs.description])
        date_type = psql_ext.new_type(oids, "Date", typecast_to_Date)
        psql_ext.register_type(date_type)

        # --- register a no-op loads() function to turn-off automatic parsing
        #     of json/jsonb datatype
        psql_extras.register_default_json(loads=lambda x: x)
        psql_extras.register_default_jsonb(loads=lambda x: x)

        # --- validation: if the following tables don't exist in the backend
        #                 sql database the client cannot work properly.
        if check:
            curs = self.conn.cursor()
            for table in self.core_tables:
                try:
                    curs.execute("SELECT 1 FROM {0:s} LIMIT 1;".format(table))
                except psql.ProgrammingError:
                    raise RuntimeError("Table {0!s} is missing on "
                                       "{1!s}".format(table, self.dbname))

    # -------------------------------------------------------------------------
    def initialize(self):
        curs = self.conn.cursor()
        curs.execute("""
-- this table is used to lookup class definition from class name
CREATE TABLE
ClassDefinitions (
    ObjType  varchar(64) PRIMARY KEY,
    ClassDef bytea       NOT NULL );

-- this table is used to store UFO objects
CREATE TABLE
Objects (
    Name        varchar(64) PRIMARY KEY,
    ObjType     varchar(64) REFERENCES ClassDefinitions,
    Version     integer     NOT NULL,
    ChangedBy   varchar(64) NOT NULL,
    TimeCreated timestamptz NOT NULL,
    LastUpdated timestamptz NOT NULL,
    Data        jsonb       NOT NULL );

CREATE INDEX objects_objtype_idx ON Objects (ObjType);

-- this table is used to store archived (i.e. dated) attributes
CREATE TABLE
Archive (
    Name      varchar(64) REFERENCES Objects,
    Attribute varchar(64) NOT NULL,
    Date      date        NOT NULL,
    Value     jsonb       NOT NULL,
    ChangedBy varchar(64) NOT NULL,
    TimeStamp timestamptz NOT NULL,
    CONSTRAINT name_attr_date UNIQUE (Name, Attribute, Date) );

CREATE INDEX archive_name_attr_idx ON Archive (Name, Attribute);

-- store procedure for deleting objects
CREATE OR REPLACE FUNCTION delete_obj(character varying, integer)
RETURNS boolean AS $$
DECLARE
    deleted record;
    remaining record;
BEGIN
    DELETE FROM Objects
    WHERE Name=$1 AND Version=$2
    RETURNING Name, ObjType INTO STRICT deleted;

    EXECUTE format('SELECT * FROM Objects
                    WHERE ObjType=$1 AND Name<>$2 LIMIT 1')
    INTO remaining
    USING deleted.ObjType, deleted.Name;

    IF NOT FOUND THEN
        EXECUTE format('DELETE FROM ClassDefinitions WHERE ObjType=$1;')
        USING deleted.ObjType;
    END IF;

    RETURN true;

EXCEPTION
    WHEN NO_DATA_FOUND THEN
        RAISE EXCEPTION 'object %, version % not found', $1, $2;
END;
$$ LANGUAGE plpgsql;""")

    # -------------------------------------------------------------------------
    def cleanup(self):
        curs = self.conn.cursor()
        curs.execute("""
DROP FUNCTION IF EXISTS delete_obj(character varying, integer) CASCADE;
DROP TABLE IF EXISTS Archive CASCADE;
DROP TABLE IF EXISTS Objects CASCADE;
DROP TABLE IF EXISTS ClassDefinitions CASCADE;""")

    # -------------------------------------------------------------------------
    def class_lookup(self, cls_name):
        try:
            return self.class_defs[cls_name]
        except KeyError:
            curs = self.conn.cursor()
            curs.execute(QUERY_CLS_GET, (cls_name,))
            row = curs.fetchone()
            if row is None:
                raise KeyError("Class definition unavailable for"
                               "{0!s} in {1!s}".format(cls_name, self.dbname))
            else:
                cls = pickle.loads(bytes(row[0]))
                self.class_defs[cls_name] = cls
                return cls

    # -------------------------------------------------------------------------
    def __contains__(self, name):
        if name in onyx_db.obj_instances:
            return True
        curs = self.conn.cursor()
        curs.execute(QUERY_OBJ_EXISTS, (name,))
        return curs.fetchone().exists

    # -------------------------------------------------------------------------
    def restart(self):
        self.close()
        self.conn = psql.connect(host=self.host,
                                 database=self.dbname, user=self.user)
        self.conn.set_isolation_level(psql_ext.ISOLATION_LEVEL_AUTOCOMMIT)

    # -------------------------------------------------------------------------
    def close(self):
        if hasattr(self, "conn") and not self.conn.closed:
            self.conn.close()

    # -------------------------------------------------------------------------
    def add(self, obj):
        # --- validate object's Name
        if not isinstance(obj.Name, str):
            if obj.Name is None:
                raise ObjDbError("Cannot persist objects with undefined Name")
            else:
                raise ObjDbError("Name attribute must be "
                                 "a string: {0!s}".format(type(obj.Name)))

        with self.conn.cursor() as curs:
            # --- if class definition is missing, add it to the
            #     ClassDefinitions table
            try:
                cls = self.class_lookup(obj.ObjType)
            except KeyError:
                cls = psql.Binary(pickle.dumps(obj.__class__, protocol=-1))
                curs.execute(QUERY_CLS_INSERT, (obj.ObjType, cls))

            # --- update time-stamp and changed-by on the object
            obj.LastUpdated = Date.now()
            obj.ChangedBy = self.user.upper()

            # --- serialize object and create record
            record = (obj.Name, obj.ObjType, obj.Version, obj.ChangedBy,
                      obj.TimeCreated, obj.LastUpdated, obj.to_json())

            try:
                curs.execute(QUERY_OBJ_INSERT, record)
            except psql.IntegrityError as err:
                if psql_err.lookup(err.pgcode) == "UNIQUE_VIOLATION":
                    raise ObjExists("Object {0!s} already in "
                                    "{1!s}".format(obj.Name, self.dbname))
                else:
                    raise ObjDbError("{0!s}. Name = "
                                     "{1!s}".format(err.message, obj.Name))

        self[obj.Name] = obj
        return obj

    # -------------------------------------------------------------------------
    def get(self, name, refresh=False):
        try:
            # --- if object must be retrieved from database raise exception and
            #     fall back
            if refresh:
                raise KeyError()
            # --- then try getting obj from database cache
            return self[name]
        except KeyError:
            # --- if not found in the db cache, get it from the database
            with self.conn.cursor() as curs:
                curs.execute(QUERY_OBJ_GET, (name,))
                row = curs.fetchone()

            if row is None:
                raise ObjNotFound("Object '{0!s}' not found "
                                  "in {1!s}".format(name, self.dbname))
            else:
                try:
                    obj = self[name]
                except KeyError:
                    cls = self.class_lookup(row.objtype)
                    self[name] = obj = cls.__new__(cls)

                for attr in SPECIAL:
                    setattr(obj, attr, getattr(row, attr.lower()))

                obj.from_json(row.data)

            return obj

    # -------------------------------------------------------------------------
    def update(self, obj):
        name = obj.Name

        # --- current version is used by the update query
        curr_ver = obj.Version

        # --- update time-stamp, changed-by and version on the object
        obj.LastUpdated = Date.now()
        obj.ChangedBy = self.user.upper()
        obj.Version += 1

        # --- serialize data and create record
        record = (obj.Version, obj.ChangedBy,
                  obj.LastUpdated, obj.to_json(), name, curr_ver)

        with self.conn.cursor() as curs:
            try:
                curs.execute(QUERY_OBJ_UPDATE, record)

                if curs.rowcount == 0:
                    # --- upadte failed: check why
                    curs.execute(QUERY_OBJ_VERSION, (name,))
                    res = curs.fetchone()

                    if res is None:
                        raise ObjNotFound("Object '{0!s}' not found in "
                                          "{1!s}".format(name, self.dbname))

                    if curr_ver < res.version:
                        msg = MSG_RELOAD.format(name, curr_ver, res.version)
                        raise ObjDbError(msg)

            except psql.IntegrityError as err:
                raise ObjDbError("{0!s}. Name = "
                                 "{1!s}".format(err.message, name))

        return obj

    # -------------------------------------------------------------------------
    def delete(self, obj):
        # --- call the delete method if available
        if hasattr(obj, "delete"):
            obj.delete()

        try:
            curs = self.conn.cursor()

            try:
                curs.execute(QUERY_OBJ_DELETE, (obj.Name, obj.Version))
            except psql.InternalError:
                # --- delete failed: check why
                curs.execute(QUERY_OBJ_VERSION, (obj.Name,))
                res = curs.fetchone()

                if res is None:
                    raise ObjNotFound("Object '{0!s}' not found "
                                      "in {1!s}".format(obj.Name, self.dbname))
                elif obj.Version < res.version:
                    msg = MSG_RELOAD.format(obj.Name, obj.Version, res.version)
                    raise ObjDbError(msg)
                else:
                    raise ObjDbError("Deletion of {0!s} failed without "
                                     "an obvious cause...".format(obj.Name))
            finally:
                curs.close()

        finally:
            del self[obj.Name]

    # -------------------------------------------------------------------------
    def set_dated(self, name, attr, date, value, overwrite=False):
        if not isinstance(name, str):
            if name is None:
                raise ObjDbError("Cannot archived attributes of "
                                 "an instance with undefined Name")
            else:
                raise ObjDbError("Name attribute must be "
                                 "a string: {0!s}".format(type(name)))

        value = json.dumps(value, cls=custom_encoder)
        changed_by = self.user.upper()
        time_stamp = Date.now()
        record_id = (name, attr, date)

        with self.conn.cursor() as curs:
            # --- first we try adding a new dated record. we create a savepoint
            #     so that if the insert fails we can rollback and try updating
            #     in-place.
            curs.execute("BEGIN;")
            curs.execute("SAVEPOINT save_dated;")
            record = (name, attr, date, value, changed_by, time_stamp)

            try:
                curs.execute(QUERY_ARC_INSERT, record)
            except psql.IntegrityError as err:
                if psql_err.lookup(err.pgcode) == "UNIQUE_VIOLATION":
                    if overwrite:
                        # --- ignore exception, rollback transaction to the set
                        #     savepoint and try updating in-place
                        curs.execute("ROLLBACK TO SAVEPOINT save_dated;")
                        record = (changed_by,
                                  time_stamp, value, name, attr, date)

                        try:
                            curs.execute(QUERY_ARC_UPDATE, record)
                        except psql.IntegrityError as err:
                            query = curs.mogrify(QUERY_ARC_UPDATE, record)
                            errmsg = "{0!s}\n{1!s}".format(err.message, query)
                            raise ObjDbError(errmsg)

                    else:
                        raise ObjDbError("Archived tuple {0!s} cannot "
                                         "be overwritten".format(record_id))
                else:
                    query = curs.mogrify(QUERY_ARC_INSERT, record)
                    errmsg = "{0!s}\n{1!s}".format(err.message, query)
                    raise ObjDbError(errmsg)

            # --- everything went fine, cleanup set cache and return
            curs.execute("RELEASE SAVEPOINT save_dated;")
            curs.execute("COMMIT;")
            self[record_id] = value, date

    # -------------------------------------------------------------------------
    def get_dated(self, name, attr, date, strict=False):
        record_id = (name, attr, date)
        try:
            # --- first try getting the value from database cache
            value, arc_date = self[record_id]
            return arc_date, json.loads(value, cls=custom_decoder)
        except KeyError:
            if strict:
                query = QUERY_ARC_GET_STRICT
                parms = (name, attr, date)
            else:
                query = QUERY_ARC_GET
                parms = (name, attr, name, attr, date)

            with self.conn.cursor() as curs:
                curs.execute(query, parms)
                row = curs.fetchone()

            if row is None:
                raise ObjNotFound("Archived tuple {0!s} not found "
                                  "in {1!s}".format(record_id, self.dbname))

            arc_date = Date.parse(row.date)

            # --- cache both the value and the corresponding archive date
            self[record_id] = row.value, arc_date

            if arc_date != date:
                # --- cache the value for the actual archive date too
                self[(name, attr, arc_date)] = row.value, arc_date

            return arc_date, json.loads(row.value, cls=custom_decoder)


###############################################################################
class ObjDbDummyClient(ObjDbBase):
    """
    A dummy client that never queries the database backend.
    """
    # -------------------------------------------------------------------------
    def restart(self):
        pass

    # -------------------------------------------------------------------------
    def close(self):
        pass

    # -------------------------------------------------------------------------
    def add(self, obj):
        if obj.Name in self:
            raise ObjExists("Object {0!s} already "
                            "in {1!s}".format(obj.Name, self.dbname))
        self[obj.Name] = obj
        return obj

    # -------------------------------------------------------------------------
    def get(self, name, refresh=False):
        if refresh:
            raise ObjDbError("get method of "
                             "ObjDbDummyClient does not support refresh")
        try:
            return self[name]
        except KeyError:
            raise ObjNotFound("Object '{0!s}' "
                              "not found in {1!s}".format(name, self.dbname))

    # -------------------------------------------------------------------------
    def update(self, obj):
        try:
            self[obj.Name] = obj
        except KeyError:
            raise ObjNotFound("Object '{0!s}' not "
                              "found in {1!s}".format(obj.Name, self.dbname))
        return obj

    # -------------------------------------------------------------------------
    def delete(self, obj):
        if hasattr(obj, "delete"):
            obj.delete()
        del self[obj.Name]

    # -------------------------------------------------------------------------
    def set_dated(self, name, attr, date, value, overwrite=False):
        self[(name, attr, date)] = value

    # -------------------------------------------------------------------------
    def get_dated(self, name, attr, date, strict=False):
        try:
            return date, self[(name, attr, date)]
        except KeyError:
            raise ObjNotFound("Archived tuple {0!s} not found in "
                              "{1!s}".format((name, attr, date), self.dbname))



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from ..datatypes.curve import Curve
from ..datatypes.gcurve import CurveError
from ..datatypes.hlocv import HlocvCurve
from ..datatypes.rdate import RDate
from .date_fns import DateRange

import numpy as np
import bisect
import collections

__all__ = [
    "CurveShift",
    "CurveIntersect",
    "CurveUnion",
    "Interpolate",
    "ApplyAdjustments",
    "ApplyToCurves",
    "CumSum",
    "Daily",
    "Weekly",
    "Monthly",
    "Quarterly",
    "Yearly",
]


# -----------------------------------------------------------------------------
def CurveShift(crv, rule, calendar=None):
    """
    Description:
        Apply the specified date rule to the each knot in the curve to obtain
        a new shifted curve.
        For a list of valid rules, see the RDate module.
    Inputs:
        crv      - the input curve
        rule     - the date-rule
        calendar - a holiday calendar used to identify business days
    Returns:
        The shifted curve.
    """
    rdt = RDate(rule, calendar)
    return crv.create_raw([d + rdt for d in crv.dates], crv.values)


# -----------------------------------------------------------------------------
def CurveIntersect(curves):
    """
    Description:
        Intersect a list of curves returning a new list of curves with only
        the common knots presents.
    Inputs:
        curves - the list of curves to be intersected
    Returns:
        A list of curves with common knots.
    """
    # --- first find dates common to all curves
    common_dates = set.intersection(*[set(crv.dates) for crv in curves])

    # --- then use these dates to build a list of curves with common knots
    intersected_curves = []
    for crv in curves:
        idxs = np.array([k for k, d in
                         enumerate(crv.dates) if d in common_dates])
        if len(idxs):
            new_crv = crv.create_raw(crv.dates[idxs], crv.values[idxs])
        else:
            new_crv = crv.create_raw([], [])
        intersected_curves.append(new_crv)

    return intersected_curves


# -----------------------------------------------------------------------------
def CurveUnion(crv1, crv2):
    """
    Description:
        Return a curve formed by the union of knots from the two input
        curves. If common knots are present, those from the first curve
        will be used.
    Inputs:
        crv1 - the first  curve (takes precedence)
        crv2 - the second curve
    Returns:
        A new curve with the union of knots.
    """
    if type(crv1) != type(crv2):
        raise CurveError("CurveUnion: curves must be of the same "
                         "type: {0!s} - {1!s}".format(type(crv1), type(crv2)))

    # --- create auxiliary dictionaries
    d1 = dict(zip(crv1.dates, crv1.values))
    d2 = dict(zip(crv2.dates, crv2.values))

    # --- merge the two dictionaries giving priority to d1
    d2.update(d1)

    # --- NB: the curve constructor sorts knots by date automatically
    return crv1.__class__(d2.keys(), d2.values(), crv1.dtype)


# -----------------------------------------------------------------------------
def Interpolate(crv, d, method="Step"):
    """
    Description:
        Return the value at curve knot using the required interpolation scheme.
    Inputs:
        crv    - the curve
        d      - the knot date (a Date)
        method - interpolation scheme. Choose among:
                    "Step":   return previous knot
                    "Linear": linear interpolation between knots
    Returns:
        The curve knot's value.
    """
    if not len(crv):
        raise CurveError("Interpolate: input curve has no data")

    try:
        return crv[d]
    except IndexError:
        if method == "Step":
            idx = bisect.bisect_left(crv.dates, d)
            if idx:
                return crv.values[idx-1]
            else:
                return crv.values[0]

        elif method == "Linear":
            idx = bisect.bisect_left(crv.dates, d)
            if idx == 0:
                return crv.values[0]
            elif idx == len(crv.dates):
                return crv.values[-1]
            else:
                d0 = crv.dates[idx-1]
                d1 = crv.dates[idx]
                w = (d - d0) / (d1 - d0)
                return w*crv.values[idx] + (1.0 - w)*crv.values[idx-1]

        else:
            raise NameError("Interpolate: unrecognized "
                            "interpolation method:{0:s}".format(method))


# -----------------------------------------------------------------------------
def ApplyAdjustments(crv, adj_crv):
    """
    Description:
        Apply a curve of multiplicative adjustments to an other curve. At the
        moment only backward looking adjustments are implemented.
    Inputs:
        crv     - the input curve
        adj_crv - the curve of adjustments
    Returns:
        A new curve of same type as crv.
    """
    if not len(crv):
        raise CurveError("ApplyAdjustments: input curve has no data")

    vls = crv.values
    dts = crv.dates

    if isinstance(crv, Curve):
        i = 0
        for d, v in adj_crv:
            j = bisect.bisect_left(dts, d)
            vls[i:j] *= v
            i = j

    elif isinstance(crv, HlocvCurve):
        i = 0
        for d, v in adj_crv:
            j = bisect.bisect_left(dts, d)
            vls[i:j,:4] *= v  # analysis:ignore
            i = j

    else:
        raise TypeError("Invalid curve type: {0!s}".format(type(crv)))

    return crv.create_raw(dts, vls)


# -----------------------------------------------------------------------------
def CumSum(crv):
    """
    Description:
        Compute the cumulative sum of the curve values.
    Inputs:
        crv - the input curve
    Returns:
        A new curve with the cumulative sum.
    """
    if not isinstance(crv, Curve):
        raise CurveError("Cumulative sum is only available "
                         "for Curve objects, not for {0!s}".format(type(crv)))

    return crv.create_raw(crv.dates, np.cumsum(crv.values))


# -----------------------------------------------------------------------------
def ApplyToCurves(curves, func=min):
    """
    Description:
        Create a new curve appling a given function to the common knots of a
        list of curves.
    Inputs:
        curves - a list of curves
        func   - the function applied to the common knots (such as min, max,
                 mean, median, etc)
    Returns:
        A new curve where every common knot has value given by func(values)
        where values are for the same date.
    """
    curves = CurveIntersect(curves)

    crv = curves.pop()
    vls = [[v] for v in crv.values]

    for crv in curves:
        for k, v in enumerate(crv.values):
            vls[k].append(v)

    return crv.create_raw(crv.dates, [func(v) for v in vls])


# -----------------------------------------------------------------------------
def Daily(crv, method="InterpolateStep", calendar=None):
    """
    Description:
        Return a daily (business days) curve obtained by up-sampling via
        interpolation/extrapolation.
    Inputs:
        crv      - the input curve
        method   - interpolation method
        calendar - a holiday calendar used to identify business days
    Returns:
        A new curve with daily knots.
    """
    dates = list(DateRange(crv.front.date, crv.back.date, "+1b", calendar))
    values = [Interpolate(crv, d, method) for d in dates]

    return crv.create_raw(dates, values)


# -----------------------------------------------------------------------------
def Weekly(crv):
    """
    Description:
        Return a weekly averaged curve. Weekly values are set on Mondays.
    Inputs:
        crv - the input curve
    Returns:
        A new curve with weekly knots.
    """
    rd = RDate("+7d")
    d0 = crv.front.date + RDate("-1M")
    d1 = d0 + rd
    avg = collections.defaultdict(list)
    for d, v in crv:
        while True:
            if d < d1:
                avg[d0].append(v)
                break
            else:
                d0 = d1
                d1 = d0 + rd

    return Curve(avg.keys(), [np.mean(v) for v in avg.values()])


# -----------------------------------------------------------------------------
def Monthly(crv):
    """
    Description:
        Return a monthly averaged curve. Monthly values are set on the first
        day of the month.
    Inputs:
        crv - the input curve
    Returns:
        A new curve with monthly knots.
    """
    rd = RDate("+1m")
    d0 = crv.front.date + RDate("+0J")
    d1 = d0 + rd
    avg = collections.defaultdict(list)
    for d, v in crv:
        while True:
            if d < d1:
                avg[d0].append(v)
                break
            else:
                d0 = d1
                d1 = d0 + rd

    return Curve(avg.keys(), [np.mean(v) for v in avg.values()])


# -----------------------------------------------------------------------------
def Quarterly(crv):
    """
    Description:
        Return a quarterly averaged curve. Quarterly average values are set on
        the first day of the quarter.
    Inputs:
        crv - the input curve
    Returns:
        A new curve with quarterly knots.
    """
    rd = RDate("+Q+1d")
    d0 = crv.front.date + RDate("+q")
    d1 = d0 + rd
    avg = collections.defaultdict(list)
    for d, v in crv:
        while True:
            if d < d1:
                avg[d0].append(v)
                break
            else:
                d0 = d1
                d1 = d0 + rd

    return Curve(avg.keys(), [np.mean(v) for v in avg.values()])


# -----------------------------------------------------------------------------
def Yearly(crv):
    """
    Description:
        Return a yearly averaged curve. Year average values are set on the
        first day of the year.
    Inputs:
        crv - the input curve
    Returns:
        A new curve with yearly knots.
    """
    rd = RDate("+E+1d")
    d0 = crv.front.date + RDate("+A")
    d1 = d0 + rd
    avg = collections.defaultdict(list)
    for d, v in crv:
        while True:
            if d < d1:
                avg[d0].append(v)
                break
            else:
                d0 = d1
                d1 = d0 + rd

    return Curve(avg.keys(), [np.mean(v) for v in avg.values()])



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from ..database.ufo_base import get_base_classes
from ..database.objdb import ObjNotFound
from ..database.objdb_api import DelObj
from .graph import DependencyGraph, SettableNode, CallableNode, create_node
from .graph import BaseVT, PropertyVT, SettableVT, CallableVT, PropSubGraphVT
from .graph import GraphError

from .. import database as onyx_db
from .. import depgraph as onyx_dg

import copy

__all__ = [
    "UseGraph",
    "ValueType",
    "CreateInMemory",
    "RemoveFromGraph",
    "PurgeObj",
    "GetNode",
    "GetVal",
    "SetVal",
    "InvalidateNode",
    "NodesByObjName",
    "ValueTypesByInstance",
    "IsInstance",
    "ChildrenSet",
    "LeafNodes",
    "SettableNodes",
]


###############################################################################
class UseGraph(object):
    """
    Context manager used to activate the dependency graph.
    """
    def __init__(self, graph=None):
        self.graph = graph or DependencyGraph()
        self.previous = None

    def __enter__(self):
        # --- make sure the global database client is available
        if onyx_db.obj_clt is None:
            raise GraphError("global database client is not "
                             "available (did you call UseDatabase?)")

        # --- swap the graph instances
        self.previous = onyx_dg.graph
        onyx_dg.graph = self.graph

    def __exit__(self, *args, **kwds):
        onyx_dg.graph = self.previous
        # --- returns False so that all execptions raised will be propagated
        return False


###############################################################################
class ValueType(object):
    """
    This decorator trasforms methods of classes derived from UfoBase into a
    ValueType descriptor.
    """
    # -------------------------------------------------------------------------
    def __init__(self, vt_type="Property", fget=None, fset=None):
        self.vt_type = vt_type
        if vt_type == "Settable":
            if fget is None or fset is None:
                raise GraphError("Settable VT needs both getter and setter")
            else:
                self.fget = fget
                self.fset = fset

    # -------------------------------------------------------------------------
    def __call__(self, func):
        if self.vt_type == "Property":
            return PropertyVT(func)
        elif self.vt_type == "Settable":
            return SettableVT(self.fget, self.fset, func.__name__)
        elif self.vt_type == "Callable":
            return CallableVT(func)
        elif self.vt_type == "PropSubGraph":
            return PropSubGraphVT(func)
        else:
            raise GraphError("Unrecognized VT type {0:s}".format(self.vt_type))


# -----------------------------------------------------------------------------
def CreateInMemory(instance):
    """
    Description:
        Helper function that adds an instance of an object derived from
        UfoBase in memory so that  all its VTs are on-graph and their values
        can be retrieved using the standard API (GetVal or graph when from
        within a mothod decorated by ValueType).
    Inputs:
        instance - the instance of an object derived from UfoBase
    Returns:
        The instance itself.
    """
    if instance.Name is None:
        raise GraphError("instance.Name was not set")

    try:
        # --- return a reference to the cached instance with same name
        return onyx_db.obj_instances[instance.Name]
    except KeyError:
        # --- add object instance to global cache
        onyx_db.obj_instances[instance.Name] = instance
        # --- return a reference to the instance
        return instance


# -----------------------------------------------------------------------------
def RemoveFromGraph(obj):
    """
    Description:
        Remove all VTs of an object from the graph.
    Inputs:
        obj - instance (or name) of an object derived from UfoBase
    Returns:
        None.
    """
    name = obj if isinstance(obj, str) else obj.Name

    # --- first invalidate all nodes referring to this object
    for node in NodesByObjName(name):
        node.invalidate()

    # --- then remove from the graph all nodes referring to this object
    for node in NodesByObjName(name):
        del onyx_dg.graph[node.get_id()]

    # --- finally remove the object itself from the VT mapping (if present)
    onyx_dg.graph.vts_by_obj.pop(name, None)


# -----------------------------------------------------------------------------
def PurgeObj(obj):
    """
    Description:
        Remove all VTs of an object from the graph and then delete the object
        from the database.
    Inputs:
        obj - an instance of an object derived from UfoBase (or its name)
    Returns:
        None.
    """
    RemoveFromGraph(obj)
    try:
        DelObj(obj)
    except ObjNotFound:
        pass


# -----------------------------------------------------------------------------
def GetNode(obj, VT, args=()):
    """
    Description:
        Fetches a node from the graph if present, otherwise it creates the new
        node and adds it to the graph.
        NB: application code shouldn't ever need to fetch a node directly, but
            rather use GetVal.
    Inputs:
        obj  - instance (or name) of an object derived from UfoBase
        VT   - name of the target ValueType
        args - tuple of arguments defining a callable node (optional)
    Returns:
        A hard reference to the node in the graph.
    """
    name = obj if isinstance(obj, str) else obj.Name
    node_id = (name, VT, args)

    try:
        return onyx_dg.graph[node_id]
    except KeyError:
        onyx_dg.graph[node_id] = node = create_node(node_id)
        return node


# -----------------------------------------------------------------------------
def GetVal(obj, VT, *args, **kwds):
    """
    Description:
        This is the proper way of calling a VT off-graph. Use this method to
        get a value type from a script.
    Inputs:
        obj    - instance (or name) of an object derived from UfoBase
        VT     - name of the target ValueType
        *args  - positional arguments used to call the target method
        **kwds -      named arguments used to call the target method
    Returns:
        The VT value.
    """
    name = obj if isinstance(obj, str) else obj.Name
    node_id = (name, VT, args)

    try:
        node = onyx_dg.graph[node_id]
    except KeyError:
        onyx_dg.graph[node_id] = node = create_node(node_id)

    # --- return node value. we make a deepcopy to make sure we never return a
    #     reference to a mutable value
    return copy.deepcopy(node.get_value(*args, **kwds))


# -----------------------------------------------------------------------------
def SetVal(obj, VT, value):
    """
    Description:
        Set the value of a stored attribute on-graph (i.e. the new value is
        visible on the graph). When the instance is persisted by calling
        UpdateObj(obj.Name), the new attribute's value is persisted as well.
        NB: only leaf nodes (i.e. nodes that are object's stored attributes)
            can be set.
    Inputs:
        obj   - instance (or name) of an object derived from UfoBase
        VT    - name of the target ValueType
        value - the value to set the VT to
    Returns:
        None.
    """
    name = obj if isinstance(obj, str) else obj.Name

    # --- get the target node from the graph
    node_id = (name, VT, ())

    try:
        node = onyx_dg.graph[node_id]
    except KeyError:
        onyx_dg.graph[node_id] = node = create_node(node_id)

    if not isinstance(node, SettableNode):
        raise GraphError("({0:s},{1:s}) "
                         "is not a settable VT".format(*node_id))

    # --- discard this node from the set of parents of all its children
    for child in node.children:
        onyx_dg.graph[child].parents.discard(node_id)

    # --- remove all its children
    node.children.clear()

    # --- invalidate all its parents
    for parent in node.parents:
        onyx_dg.graph[parent].invalidate()

    # --- set the node value and set its state to valid
    node.value = value
    node.valid = True

    # --- set attribute in the cached instance of the object
    setattr(node.obj_ref(), node.VT, node.value)


# -----------------------------------------------------------------------------
def InvalidateNode(obj, VT, args=()):
    """
    Description:
        Invalidate a node in the graph.
    Inputs:
        obj  - instance (or name) of an object derived from UfoBase
        VT   - name of the target ValueType
        args - tuple of arguments defining a callable node (optional)
    Returns:
        None.
    """
    name = obj if isinstance(obj, str) else obj.Name
    node = GetNode(name, VT, args)
    if isinstance(node, CallableNode) and args == ():
        raise GraphError("To invalidate a clalable node it is "
                         "mandatory to specify the tuple of arguments")
    node.invalidate()


# -----------------------------------------------------------------------------
def NodesByObjName(obj):
    """
    Description:
        Generator: returns all the nodes that reference the same object
    Inputs:
        obj - instance (or name) of an object derived from UfoBase
    Yields:
        Graph nodes.
    """
    name = obj if isinstance(obj, str) else obj.Name
    try:
        for VT, args in onyx_dg.graph.vts_by_obj[name]:
            yield onyx_dg.graph[(name, VT, args)]
    except KeyError:
        raise StopIteration


# -----------------------------------------------------------------------------
def ValueTypesByInstance(obj):
    """
    Description:
        Returns a list of all value types for object obj
    Inputs:
        obj - instance (or name) of an object derived from UfoBase
    Returns:
        A list of VT names.
    """
    if isinstance(obj, str):
        obj = onyx_db.obj_clt.get(obj)

    cls = obj.__class__
    calc = [name for name, VT in
            cls.__dict__.items() if isinstance(VT, BaseVT)]

    return list(obj.StoredAttrs) + calc


# -----------------------------------------------------------------------------
def IsInstance(name, obj_type):
    """
    Description:
        This is the equivalent of isinstance, but takes the instance name and
        the class name as inputs.
    Inputs:
        name     - the instance name
        obj_type - the class name
    """
    obj = onyx_db.obj_clt.get(name)
    bases = set(get_base_classes(obj.__class__))
    return obj_type in bases


# -----------------------------------------------------------------------------
def children_iterator(node_id, child_VT, obj_type):
    for name, vt, args in onyx_dg.graph[node_id].children:
        if vt == child_VT:
            if obj_type is None:
                yield name
            else:
                obj = onyx_db.obj_instances[name]
                bases = set(get_base_classes(obj.__class__))
                if obj_type in bases:
                    yield name
        else:
            for kid in children_iterator((name, vt, args), child_VT, obj_type):
                yield kid


# -----------------------------------------------------------------------------
def ChildrenSet(node_id, child_VT, obj_type=None, graph=None):
    """
    Description:
        Given a node id, return the names of all the objects that, combined
        with a specific VT, are its children.
    Inputs:
        node_id  - parent node_id, in the form (Name, VT, args), where args can
                   be skipped in place of an empty tuple
        child_VT - children ValueType's name
        obj_type - if set, children must be subclass of this class
        graph    - if set, function used to generate graph topology
    Yields:
        A set with object names.
    """
    obj, VT, *args = node_id
    name = obj if isinstance(obj, str) else obj.Name
    node_id = (name, VT, tuple(args))

    if graph is None:
        GetVal(obj, VT, *args)
    else:
        graph(obj, VT, *args)

    return set(children_iterator(node_id, child_VT, obj_type))


# -----------------------------------------------------------------------------
def leaves_iterator(node_id, get_val):
    stored = get_val(node_id[0], "StoredAttrs")

    for name, vt, args in onyx_dg.graph[node_id].children:
        if vt in stored:
            yield name, vt
        else:
            for leaf in leaves_iterator((name, vt, args), get_val):
                yield leaf


# -----------------------------------------------------------------------------
def LeafNodes(node_id, graph=None):
    """
    Description:
        Return all leaf-level nodes (StoredAttrs) of a given node.
    Inputs:
        node_id - parent node_id, in the form (Name, VT, args), where args can
                  be skipped in place of an empty tuple
        graph   - if set, function used to generate graph topology
    Yields:
        A set of nodes.
    """
    obj, VT, *args = node_id
    name = obj if isinstance(obj, str) else obj.Name
    node_id = (name, VT, tuple(args))

    get_val = graph or GetVal
    get_val(obj, VT, *args)

    return set(leaves_iterator(node_id, get_val))


# -----------------------------------------------------------------------------
def settable_iterator(node_id):
    for kid_id in onyx_dg.graph[node_id].children:
        node = onyx_dg.graph[kid_id]
        for leaf in settable_iterator(kid_id):
            if leaf is not None:
                yield leaf

        if isinstance(node, SettableNode):
            yield kid_id


# -----------------------------------------------------------------------------
def SettableNodes(node_id, graph=None):
    """
    Description:
        Return all settable children of a given node (stopping at the first
        level).
    Inputs:
        node_id - parent node_id, in the form (Name, VT, args), where args can
                  be skipped in place of an empty tuple
        graph   - if set, function used to generate graph topology
    Yields:
        A set of nodes.
    """
    obj, VT, *args = node_id
    name = obj if isinstance(obj, str) else obj.Name
    node_id = (name, VT, tuple(args))

    if graph is None:
        GetVal(obj, VT, *args)
    else:
        graph(obj, VT, *args)

    return set(settable_iterator(node_id))



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core.database.ufo_base import UfoBase
from onyx.core.database.ufo_fields import IntField
from onyx.core.depgraph.graph_api import ValueType, CreateInMemory
from onyx.core.depgraph.graph_api import GetVal, SetVal, GetNode
from onyx.core.depgraph.graph_scopes import EvalBlock, GraphScope
from onyx.core.utils.unittest import OnyxTestCase

import unittest

FIXED = 1


###############################################################################
class UfoClosure(UfoBase):
    fixed = IntField(default=FIXED)

    # -------------------------------------------------------------------------
    @ValueType()
    def add_to_fixed(self, graph):
        def adder(x):
            return x + graph(self, "fixed")
        return adder


###############################################################################
class TestClosure(OnyxTestCase):
    # -------------------------------------------------------------------------
    def setUp(self):
        super().setUp()
        self.instance = CreateInMemory(UfoClosure(Name="closure"))
        self.name = self.instance.Name

    # -------------------------------------------------------------------------
    def test_closure(self):
        self.assertEqual(GetVal(self.name, "fixed"), FIXED)
        self.assertEqual(GetVal(self.name, "add_to_fixed")(10), 10 + FIXED)

        # --- inspect the node before and after changing the value of "fixed"
        adder = GetNode(self.name, "add_to_fixed")
        self.assertTrue(adder.valid)

        SetVal(self.name, "fixed", 666)
        self.assertFalse(adder.valid)

        self.assertEqual(GetVal(self.name, "add_to_fixed")(10), 10 + 666)
        self.assertTrue(adder.valid)

    # -------------------------------------------------------------------------
    def test_invalidation_within_EvalBlock(self):
        self.assertEqual(GetVal(self.name, "add_to_fixed")(10), 10 + FIXED)
        adder = GetNode(self.name, "add_to_fixed")
        self.assertTrue(adder.valid)

        with EvalBlock() as eb:
            self.assertTrue(adder.valid)
            eb.change_value(self.name, "fixed", 666)
            self.assertFalse(adder.valid)

            self.assertEqual(GetVal(self.name, "add_to_fixed")(10), 10 + 666)
            self.assertTrue(adder.valid)

        self.assertFalse(adder.valid)

    # -------------------------------------------------------------------------
    def test_invalidation_within_GraphScope(self):
        self.assertEqual(GetVal(self.name, "add_to_fixed")(10), 10 + FIXED)
        outer_adder = GetNode(self.name, "add_to_fixed")
        self.assertTrue(outer_adder.valid)

        with GraphScope() as scope:
            # --- fetch node again from within the graph scope: this the same
            #     as the outer node
            inner_adder = GetNode(self.name, "add_to_fixed")

            self.assertEqual(id(inner_adder), id(outer_adder))
            self.assertTrue(outer_adder.valid)
            self.assertTrue(inner_adder.valid)
            self.assertEqual(inner_adder.value, outer_adder.value)

            # --- we now change the graph from withing a scope and the two
            #     nodes start living separate lives
            scope.change_value(self.name, "fixed", 666)

            inner_adder = GetNode(self.name, "add_to_fixed")

            self.assertTrue(id(inner_adder) != id(outer_adder))
            self.assertTrue(outer_adder.valid)
            self.assertFalse(inner_adder.valid)

            self.assertEqual(GetVal(self.name, "add_to_fixed")(10), 10 + 666)

        self.assertTrue(outer_adder.valid)
        self.assertEqual(GetVal(self.name, "add_to_fixed")(10), 10 + FIXED)


if __name__ == "__main__":
    from onyx.core.utils.unittest import UseEphemeralDbs
    with UseEphemeralDbs():
        unittest.main(failfast=True)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core.datatypes.date import Date
from onyx.core.database.objdb import ObjDbDummyClient, ObjNotFound
from onyx.core.database.test import ufo_testcls

import unittest


# --- Unit tests
class RegTest(unittest.TestCase):
    def test_add_get_update_delete(self):
        clt = ObjDbDummyClient("dummy_db", "dummy_user")
        obj = ufo_testcls.ufocls(Name="test")

        self.assertRaises(ObjNotFound, clt.get, obj.Name)
        self.assertEqual(clt.add(obj), obj)
        self.assertEqual(clt.get(obj.Name), obj)

        obj.Birthday = Date.today()
        self.assertEqual(clt.update(obj), obj)

        clt.delete(obj)
        self.assertRaises(ObjNotFound, clt.get, obj.Name)

if __name__ == "__main__":
    unittest.main()



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from calendar import SATURDAY, SUNDAY

__all__ = ["HolidayCalendar"]


###############################################################################
class HolidayCalendar(object):
    """
    Store holidays and provide testing methods.
    """
    # -------------------------------------------------------------------------
    def __init__(self, holidays=None):
        if holidays is None:
            self.holidays = set()
        else:
            self.holidays = set(holidays)

    # -------------------------------------------------------------------------
    def add(self, holidays):
        for holiday in holidays:
            self.holidays.add(holiday)

    # -------------------------------------------------------------------------
    def is_holiday(self, d):
        if d in self.holidays:
            return True
        elif d.weekday() in (SATURDAY, SUNDAY):
            # --- saturday and sundays are always recurring holidays
            return True
        return False

    # -------------------------------------------------------------------------
    def __deepcopy__(self, memo):
        clone = self.__class__.__new__(self.__class__)
        clone.holidays = self.holidays.copy()
        memo[id(self)] = clone
        return clone



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

# FIXME: distutils problem with hardlinks
import os
del os.link

from setuptools import setup, find_packages

install_requires = [
    "dateutils",
    "numpy",
    "psycopg2",
    "argh",
]

DESCRIPTION = """
Onyx is a framework to develop integrated trading, position management, pricing
and risk management platforms.
onyx.core implements the fundamental datatypes, various serialization tools,
and a powerful dependency graph."""

setup(
    name="onyx.core",
    setup_requires=["hgtools"],
    use_vcs_version={"increment": "0.1"},
    description="Core functions and datatypes for the onyx framework.",
    long_description=DESCRIPTION,
    author="carlo sbraccia",
    author_email="carlosbraccia@yahoo.co.uk",
    url="https://bitbucket.org/sbraccia/onyx.core",
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    entry_points={
        "console_scripts": [
            "run-forever = onyx.core.utils.run_forever:main",
        ],
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Financial and Insurance Industry",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.4",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
    ],
)
